@@ -1,18 +1,11 @@
 package org.perl6.nqp.io;
 
 import java.io.BufferedReader;
-import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.InputStreamReader;
 import java.io.OutputStreamWriter;
-import java.nio.ByteBuffer;
-import java.nio.CharBuffer;
-import java.nio.channels.FileChannel;
 import java.nio.charset.Charset;
-import java.nio.file.Path;
-import java.nio.file.StandardOpenOption;
-import java.util.ArrayList;
 
 import jline.ConsoleReader;
 
@@ -43,6 +36,22 @@ public class StandardReadHandle implements IIOClosable, IIOEncodable, IIOSyncRea
         this.cs = cs;
     }
     
+    public byte[] read(ThreadContext tc, int bytes) {
+        try {
+            byte[] array = new byte[bytes];
+            int read = 0;
+            int offset = 0;
+            while ((read = is.read(array, offset, bytes - offset)) != -1) {
+                offset += read;
+            }
+            byte[] compact = new byte[offset];
+            System.arraycopy(array, 0, compact, 0, offset);
+            return compact;
+        } catch (IOException e) {
+            throw ExceptionHandling.dieInternal(tc, e);
+        }
+    }
+
     public synchronized String slurp(ThreadContext tc) {
         try {
             if (br == null)
@@ -74,6 +83,23 @@ public class StandardReadHandle implements IIOClosable, IIOEncodable, IIOSyncRea
         }
     }
     
+    public synchronized String getc(ThreadContext tc) {
+        try {
+            if (br == null)
+                br = new BufferedReader(new InputStreamReader(is, cs));
+            int read = br.read();
+            if (read == -1) {
+                eof = true;
+                return "";
+            }
+            else {
+                return String.valueOf(read);
+            }
+        } catch (IOException e) {
+            throw ExceptionHandling.dieInternal(tc, e);
+        }
+    }
+    
     public synchronized String readlineInteractive(ThreadContext tc, String prompt) {
         try {
             if (cr == null)
