@@ -14,6 +14,7 @@ import org.objectweb.asm.Opcodes;
 import org.objectweb.asm.Type;
 import org.perl6.nqp.runtime.ExceptionHandling;
 import org.perl6.nqp.runtime.ThreadContext;
+import org.perl6.nqp.sixmodel.ByteClassLoader;
 import org.perl6.nqp.sixmodel.REPR;
 import org.perl6.nqp.sixmodel.STable;
 import org.perl6.nqp.sixmodel.SerializationReader;
@@ -658,14 +659,15 @@ public class P6Opaque extends REPR {
         // If there's a different number of attributes, need to set up delegate.
         // Note the condition below works because we don't make an entry in the
         // class handles list for a type with no attributes.
+        P6OpaqueBaseInstance instance = (P6OpaqueBaseInstance)obj;
         if (ourREPRData.classHandles.length != targetREPRData.classHandles.length) {
             // Create delegate.
             SixModelObject delegate = newType.st.REPR.allocate(tc, newType.st);
             
             // Find original object.
             SixModelObject orig;
-            if (((P6OpaqueBaseInstance)obj).delegate != null)
-                orig = ((P6OpaqueBaseInstance)obj).delegate;
+            if (instance.delegate != null)
+                orig = instance.delegate;
             else
                 orig = obj;
             
@@ -679,7 +681,13 @@ public class P6Opaque extends REPR {
             catch (IllegalAccessException e) { throw new RuntimeException(e); }
 
             // Install.
-            ((P6OpaqueBaseInstance)obj).delegate = delegate;
+            instance.delegate = delegate;
+        }
+        // If we don't have to create a new delegate, but there's a delegate
+        // in place already, the delegate needs a new STable as well.
+        // Otherwise, methods introduced in the new type won't be available.
+        else if(instance.delegate != null) {
+            instance.delegate.st = newType.st;
         }
         
         // Switch STable over to the new type.
@@ -712,18 +720,6 @@ public class P6Opaque extends REPR {
         return STable.NO_HINT;
     }
     
-    private class ByteClassLoader extends ClassLoader {
-        private byte[] bytes;
-        
-        public ByteClassLoader(byte[] bytes) {
-            this.bytes = bytes;
-        }
-        
-        public Class<?> findClass(String name) {
-            return defineClass(name, this.bytes, 0, this.bytes.length);
-        }
-    }
-    
     @SuppressWarnings("unchecked")
     public void deserialize_repr_data(ThreadContext tc, STable st, SerializationReader reader) {
         // Instantiate REPR data.
